#-----------------------
# @author: Tony Ribeiro
# @date: 2018/06/02
#
# ILP2018 evaluation of GULA
# 
#-----------------------

from __future__ import print_function
import sys
from subprocess import call
import math
import random
import datetime
import time

#-----------------------
# Stderr print function
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
#-----------------------

#-----------------------
# rounding function
def roundup(x):
    return int(math.ceil(x / 10.0)) * 10
#-----------------------

# 0) Parameters management
#--------------------------

eprint("> Starting System generation...")
eprint(">> Parameter extraction...")

random.seed(a=None)

#******************
# parameters      *
#
experiment_id = "" 		# $1
semantic = "general" 	# $2
timeout = 10000
#                 *
#******************

if len(sys.argv) < 2:
	eprint("No experiement specified must be:\n - 1: correctness\n - 2: scalability over variables\n - 3: scalability over transitions\n")
	sys.exit()

experiment_id = int(sys.argv[1])
eprint("Experiment ID given: ",experiment_id)

if len(sys.argv) > 2:
	semantic = sys.argv[2]
	if semantic != "synchronous" and semantic != "asynchronous" and semantic != "general":
		eprint("Parameter error: semantic must be one of synchronous, asynchronous or general")
		sys.exit()
	eprint("Semantic given: ",semantic)
else:
	eprint("Default semantic: "+semantic)

eprint ("> Starting ILP2018 GENLF1T Evaluation...")
start_time = time.time()

f = open("temp/full_output.txt", "w")
f.close()

# 1) Correctness check
#-----------------------------------
if experiment_id == 1:
	min_var = 1
	max_var = 10
	step = 1
	nb_levels = 2
	nb_instances = 1
	
	eprint (">> Starting correctness experiments: transitions regereration...")
	
	for nb_var in range(min_var,max_var,step):
		eprint(">>> System size: "+str(nb_var))
		sum_run_time = 0
		
		for instance_id in range(0,nb_instances):
			eprint(">>>> Instance: "+str(instance_id))
			instance_start_time = time.time()
			
			# 1.1) Generate all transitions of a random system 
			eprint(">>>>> Generating training transitions...")
			instance_file_path = "temp/var_"+str(nb_var)+"_id_"+str(instance_id)+"_training_transitions"
			out = open(instance_file_path, "w")
			full_out = open("temp/full_output.txt", "a")
			call(["python","programs/system_generator/main.py",str(nb_var),str(nb_levels),str(semantic)], stdout=out,stderr=full_out)
			out.close()
			full_out.close()
			
			# 1.2) Learn the system from the transitions
			eprint(">>>>> Learning from training transitions...")
			full_out = open("temp/full_output.txt", "a")
			model_file_path = "temp/var_"+str(nb_var)+"_id_"+str(instance_id)+"_GENLF1T_output"
			out = open(model_file_path,"w")
			call(["/usr/bin/time","-f","%e","-o","temp/run_time.txt","timeout",str(timeout),"programs/GEN_LF1T.exe","0",str(instance_file_path)],stdout=out,stderr=full_out)
			full_out.close()
			out.close()
			
			# 1.3) Generate transitions from the model learned
			eprint(">>>>> Generating predicted transitions...")
			predicted_transitions_file_path = "temp/var_"+str(nb_var)+"_id_"+str(instance_id)+"_predicted_transitions"
			out = open(predicted_transitions_file_path,"w")
			full_out = open("temp/full_output.txt", "a")
			call(["python","programs/transitions_generator/main.py",model_file_path,semantic],stdout=out,stderr=full_out)
			out.close()
			full_out.close()
			
			eprint(">>>>> Comparing training V.S. predicted...")
			
			# 1.4) Check missing transitions
			out = open("temp/difference.txt","w")
			call(["grep","-Fxv","-f",instance_file_path,predicted_transitions_file_path],stdout=out)
			out.close()
			
			out = open("temp/difference.txt","r")
			line = out.read()
			
			if len(line) > 0:
				eprint(">>>>>> ERROR: missing transition")
				eprint(line)
				sys.exit()	
				
			eprint(">>>>>> No missing transitions")
			
			# 1.5) Check unexpected transitions
			out = open("temp/difference.txt","w")
			call(["grep","-Fxv","-f",predicted_transitions_file_path,instance_file_path],stdout=out)
			out.close()
			
			out = open("temp/difference.txt","r")
			line = out.read()
			
			if len(line) > 0:
				eprint(">>>>>> ERROR: unexpected transition")
				eprint(line)
				sys.exit()
				
			eprint(">>>>>> No unexpected transitions")
			
			eprint(">>>> Instance: "+str(instance_id)+" OK")
			eprint("--- Local %.2f seconds ---" % (time.time() - instance_start_time))
			eprint("--- Global %.2f seconds ---" % (time.time() - start_time))
		eprint(">>> All instances OK")
	eprint(">> Correctness test OK")
	eprint("--- Global %.2f seconds ---" % (time.time() - start_time))
	
# 2) Scalability evolving variables
#-----------------------------------
if experiment_id == 2:
	min_var = 1
	max_var = 20
	step = 1
	nb_levels = 3
	nb_instances = 10
	
	eprint (">> Starting scalability experiments: varying number of variables...")
	
	f = open("results/results_run_time_variables_"+str(semantic)+".txt","w")
	f.write("#"+str(datetime.datetime.now())+"\n")
	f.write("# Scalability experiment varying number of variables:\n")
	f.write("# Semantic "+str(semantic)+"\n")
	f.write("# Size "+str(min_var)+" to "+str(max_var)+"\n")
	f.write("# Levels "+str(nb_levels)+"\n")
	f.write("# Instances "+str(nb_instances)+"\n")
	f.write("# Timeout "+str(timeout)+" seconds\n")
	f.write("\n")
	f.write("Variables, Run time\n")
	f.write("")
	f.close()

	for nb_var in range(min_var,max_var,step):
		eprint(">>> System size: "+str(nb_var))
		sum_run_time = 0
		
		# 2.1) Generate and learn instances
		for instance_id in range(0,nb_instances):
			eprint(">>>> Instance: "+str(instance_id))
			
			# 2.1.1) Generate all transitions of a random system 
			instance_file_path = "temp/var_"+str(nb_var)+"_id_"+str(instance_id)
			out = open(instance_file_path, "w")
			full_out = open("temp/full_output.txt", "a")
			call(["python","programs/system_generator/main.py",str(nb_var),str(nb_levels),str(semantic)], stdout=out,stderr=full_out)
			out.close()
			full_out.close()
			
			# 2.1.2) Learn the system from the transitions
			full_out = open("temp/full_output.txt", "a")
			call(["/usr/bin/time","-f","%e","-o","temp/run_time.txt","timeout",str(timeout),"programs/GEN_LF1T.exe","0",str(instance_file_path)],stdout=full_out,stderr=full_out)
			full_out.close()
			
			# 2.1.3) Extract run time
			f = open("temp/run_time.txt","r")
			local_run_time = f.read()
			try:
				float(local_run_time)
			except ValueError:
				eprint("Time Out reached")
				f.close()
				f = open("results/results_run_time_variables_"+str(semantic)+".txt","a")
				f.write(str(nb_var)+", T.O.\n")
				f.close()
				sys.exit()
				
			f.close()
			local_run_time = float(local_run_time)
			sum_run_time += local_run_time
			
			eprint(">>>> Runtime: "+str(local_run_time)+"s")
		
		# 2.2) Compute average run time
		avg_run_time = sum_run_time / float(nb_instances)
		eprint(">>> AVG Runtime: "+str(avg_run_time)+"s")
		
		# 2.3) Store run time
		f = open("results/results_run_time_variables_"+str(semantic)+".txt","a")
		f.write(str(nb_var)+", "+str(avg_run_time)+"\n")
		f.close()
	
	eprint(">> Scalability experiment varying number of variables finished successfully")
	eprint("--- %.2f seconds ---" % (time.time() - start_time))
	
# 3) Scalability evolving variables domains
#-------------------------------------
if experiment_id == 3:
	nb_var = 3
	min_levels = 2
	max_levels = 20
	step = 1
	nb_instances = 10
	
	eprint (">> Starting scalability experiments: varying number of variables levels...")
	
	f = open("results/results_run_time_levels_"+str(semantic)+".txt","w")
	f.write("#"+str(datetime.datetime.now())+"\n")
	f.write("# Scalability experiment varying number of variables levels:\n")
	f.write("# Semantic "+str(semantic)+"\n")
	f.write("# Size "+str(nb_var)+"\n")
	f.write("# Levels "+str(min_levels)+"to"+str(max_levels)+"\n")
	f.write("# Instances "+str(nb_instances)+"\n")
	f.write("# Timeout "+str(timeout)+" seconds\n")
	f.write("\n")
	f.write("Levels, Run time\n")
	f.write("")
	f.close()

	for nb_levels in range(min_levels,max_levels):
		eprint(">>> Domain size: "+str(nb_levels))
		sum_run_time = 0
		
		# 2.1) Generate and learn instances
		for instance_id in range(0,nb_instances):
			eprint(">>>> Instance: "+str(instance_id))
			
			# 2.1.1) Generate all transitions of a random system 
			instance_file_path = "temp/levels_"+str(nb_levels)+"_id_"+str(instance_id)
			out = open(instance_file_path, "w")
			full_out = open("temp/full_output.txt", "a")
			call(["python","programs/system_generator/main.py",str(nb_var),str(nb_levels),str(semantic)], stdout=out,stderr=full_out)
			out.close()
			full_out.close()
			
			# 2.1.2) Learn the system from the transitions
			full_out = open("temp/full_output.txt", "a")
			call(["/usr/bin/time","-f","%e","-o","temp/run_time.txt","timeout",str(timeout),"programs/GEN_LF1T.exe","0",str(instance_file_path)],stdout=full_out,stderr=full_out)
			full_out.close()
			
			# 2.1.3) Extract run time
			f = open("temp/run_time.txt","r")
			local_run_time = f.read()
			try:
				float(local_run_time)
			except ValueError:
				eprint("Time Out reached")
				f.close()
				f = open("results/results_run_time_levels_"+str(semantic)+".txt","a")
				f.write(str(nb_levels)+", T.O.\n")
				f.close()
				sys.exit()
				
			f.close()
			local_run_time = float(local_run_time)
			sum_run_time += local_run_time
			
			eprint(">>>> Runtime: "+str(local_run_time)+"s")
		
		# 2.2) Compute average run time
		avg_run_time = sum_run_time / float(nb_instances)
		eprint(">>> AVG Runtime: "+str(avg_run_time)+"s")
		
		# 2.3) Store run time
		f = open("results/results_run_time_levels_"+str(semantic)+".txt","a")
		f.write(str(nb_levels)+", "+str(avg_run_time)+"\n")
		f.close()
		
	eprint(">> Scalability experiment varying number of levels finished successfully")
	eprint("--- %.2f seconds ---" % (time.time() - start_time))
	
# 4) Scalability evolving transitions
#-------------------------------------
if experiment_id == 4:
	nb_var = 10
	min_trans = 10
	max_trans = 1000000
	step = 2
	nb_levels = 3
	nb_instances = 10
	
	eprint (">> Starting scalability experiments: varying number of transitions...")
	
	f = open("results/results_run_time_transitions.txt","w")
	f.write("#"+str(datetime.datetime.now())+"\n")
	f.write("# Scalability experiment varying number of variables:\n")
	f.write("# Semantic "+str(semantic)+"\n")
	f.write("# Size "+str(nb_var)+"\n")
	f.write("# Levels "+str(nb_levels)+"\n")
	f.write("# Instances "+str(nb_instances)+"\n")
	f.write("# Timeout "+str(timeout)+" seconds\n")
	f.write("\n")
	f.write("Transitions, Run time\n")
	f.write("")
	f.close()

	nb_trans = int(min_trans)
	while nb_trans <= max_trans:
		eprint(">>> Input size: "+str(nb_trans))
		sum_run_time = 0
		
		# 2.1) Generate and learn instances
		for instance_id in range(0,nb_instances):
			eprint(">>>> Instance: "+str(instance_id))
			
			# 2.1.1) Generate all transitions of a random system 
			instance_file_path = "temp/trans_"+str(nb_trans)+"_id_"+str(instance_id)
			out = open(instance_file_path, "w")
			full_out = open("temp/full_output.txt", "a")
			call(["python","programs/system_generator/main.py",str(nb_var),str(nb_levels),str(semantic),str(nb_trans)], stdout=out,stderr=full_out)
			out.close()
			full_out.close()
			
			# 2.1.2) Learn the system from the transitions
			full_out = open("temp/full_output.txt", "a")
			call(["/usr/bin/time","-f","%e","-o","temp/run_time.txt","timeout",str(timeout),"programs/GEN_LF1T.exe","0",str(instance_file_path)],stdout=full_out,stderr=full_out)
			full_out.close()
			
			# 2.1.3) Extract run time
			f = open("temp/run_time.txt","r")
			local_run_time = f.read()
			try:
				float(local_run_time)
			except ValueError:
				eprint("Time Out reached")
				f.close()
				f = open("results/results_run_time_transitions_"+str(semantic)+".txt","a")
				f.write(str(nb_trans)+", T.O.\n")
				f.close()
				sys.exit()
				
			f.close()
			local_run_time = float(local_run_time)
			sum_run_time += local_run_time
			
			eprint(">>>> Runtime: "+str(local_run_time)+"s")
		
		# 2.2) Compute average run time
		avg_run_time = sum_run_time / float(nb_instances)
		eprint(">>> AVG Runtime: "+str(avg_run_time)+"s")
		
		# 2.3) Store run time
		f = open("results/results_run_time_transitions_"+str(semantic)+".txt","a")
		f.write(str(nb_trans)+", "+str(avg_run_time)+"\n")
		f.close()
		
		nb_trans *= step
	eprint(">> Scalability experiment varying number of transitions finished successfully")
	eprint("--- %.2f seconds ---" % (time.time() - start_time))
	
# 5) Scalability on benchmarks
#-----------------------------------
if experiment_id == 5:
	benchmarks = ["repressilator","fission","mammalian","budding","arabidopsis"]
	timeout = 36000
	
	eprint (">> Starting scalability experiments: Boolean networks benchmarks...")
	
	f = open("results/results_run_time_benchmarks_"+str(semantic)+".txt","w")
	f.write("#"+str(datetime.datetime.now())+"\n")
	f.write("# Scalability experiment on benchmarks:\n")
	f.write("# Semantic "+str(semantic)+"\n")
	f.write("# Timeout "+str(timeout)+" seconds\n")
	f.write("\n")
	f.write("Benchmark, Run time\n")
	f.write("")
	f.close()

	for benchmark in benchmarks:
		eprint(">>> System: "+str(benchmark))
		
		
		# 2.1.1) Generate all transitions of the system 
		eprint(">>>>> Generating transitions...")
		model_file_path = "benchmarks/"+str(benchmark)+".lp"
		instance_file_path = "temp/"+str(benchmark)+"_transitions"
		out = open(instance_file_path,"w")
		full_out = open("temp/full_output.txt", "a")
		call(["python","programs/transitions_generator/main.py",model_file_path,semantic],stdout=out,stderr=full_out)
		out.close()
		full_out.close()
		
		# 2.1.2) Learn the system from the transitions
		full_out = open("temp/full_output.txt", "a")
		call(["/usr/bin/time","-f","%e","-o","temp/run_time.txt","timeout",str(timeout),"programs/GEN_LF1T.exe","0",str(instance_file_path)],stdout=full_out,stderr=full_out)
		full_out.close()
		
		# 2.1.3) Extract run time
		f = open("temp/run_time.txt","r")
		local_run_time = f.read()
		try:
			float(local_run_time)
		except ValueError:
			eprint("Time Out reached")
			f.close()
			f = open("results/results_run_time_benchmarks_"+str(semantic)+".txt","a")
			f.write(str(benchmark)+", T.O.\n")
			f.close()
			sys.exit()
			
		f.close()
		local_run_time = float(local_run_time)
		
		eprint(">>>> Runtime: "+str(local_run_time)+"s")
		
		# 2.3) Store run time
		f = open("results/results_run_time_benchmarks_"+str(semantic)+".txt","a")
		f.write(str(benchmark)+", "+str(local_run_time)+"\n")
		f.close()
	
	eprint(">> Scalability experiment over benchmarks finished successfully")
	eprint("--- %.2f seconds ---" % (time.time() - start_time))
	
# 6) (generation only) Scalability on benchmarks
#-----------------------------------
if experiment_id == 6:
	benchmarks = ["repressilator","fission","mammalian","budding","arabidopsis"]
	timeout = 36000
	
	eprint (">> Starting scalability experiments: Boolean networks benchmarks...")
	
	f = open("results/results_run_time_benchmarks_"+str(semantic)+".txt","w")
	f.write("#"+str(datetime.datetime.now())+"\n")
	f.write("# Scalability experiment on benchmarks:\n")
	f.write("# Semantic "+str(semantic)+"\n")
	f.write("# Timeout "+str(timeout)+" seconds\n")
	f.write("\n")
	f.write("Benchmark, Run time\n")
	f.write("")
	f.close()

	for benchmark in benchmarks:
		eprint(">>> System: "+str(benchmark))
		
		
		# 2.1.1) Generate all transitions of the system 
		eprint(">>>>> Generating transitions...")
		model_file_path = "benchmarks/"+str(benchmark)+".lp"
		instance_file_path = "temp/"+str(benchmark)+"_transitions"
		out = open(instance_file_path,"w")
		full_out = open("temp/full_output.txt", "a")
		call(["python","programs/transitions_generator/main.py",model_file_path,semantic],stdout=out,stderr=full_out)
		out.close()
		full_out.close()
		
		
	
	eprint(">> (generation only) Scalability experiment over benchmarks finished successfully")
